<?php

namespace Drupal\semantic_search\Services;

use Drupal\Core\Config\ConfigFactory;
use Drupal\Core\Logger\LoggerChannelFactory;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Service to calculate embeddings.
 *
 * @package Drupal\semantic_search\Services
 */
class EmbeddingVectorizationService {

  /**
   * The storage options and vector sizes.
   */
  const STORAGE_OPTIONS = [
    'openai' => [
      'key' => 'openai',
      'label' => 'Open AI',
      'vector_size' => 1536,
    ],
    'custom' => [
      'key' => 'custom',
      'label' => 'Custom',
      'vector_size' => 768,
    ],
  ];

  /**
   * The OpenAI storage key.
   */
  const OPENAI_KEY = 'openai';

  /**
   * The custom API key.
   */
  const CUSTOM_API_KEY = 'custom';

  /**
   * The Config Factory.
   *
   * @var \Drupal\Core\Config\ConfigFactory
   */
  protected ConfigFactory $configFactory;

  /**
   * The Logger Channel Factory.
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactory
   */
  protected LoggerChannelFactory $loggerFactory;

  /**
   * Constructs for EmbeddingVectorizationService.
   *
   * @param \Drupal\Core\Config\ConfigFactory $config_factory
   *   The config factory.
   * @param \Drupal\Core\Logger\LoggerChannelFactory $logger_factory
   *   The logger factory.
   */
  public function __construct(ConfigFactory $config_factory, LoggerChannelFactory $logger_factory) {
    $this->configFactory = $config_factory;
    $this->loggerFactory = $logger_factory;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('logger.factory')
    );
  }

  /**
   * Get embedding.
   *
   * @return array
   *   OpenAI embedding.
   */
  public function getEmbedding($source) {
    switch ($this->getVectorizationService()) {
      case self::OPENAI_KEY:
        $openai_key = $this->configFactory->get('semantic_search.openai_client')->get('openai_key');
        $client = \OpenAI::client($openai_key);
        $response = $client->embeddings()->create([
          'model' => 'text-embedding-ada-002',
          'input' => strip_tags($source),
        ]);

        foreach ($response->embeddings as $embedding) {
          $embedding = $embedding->embedding;
        }
        break;

      case self::CUSTOM_API_KEY:

        // Generated by curl-to-PHP: http://incarnate.github.io/curl-to-php/
        $ch = curl_init();

        $custom_vectorization_api_config = $this->configFactory->get('semantic_search.custom_vectorization_api');

        $search_string = str_replace('"', "", $source);
        $search_string = str_replace(["\r", "\n"], '', $search_string);
        $search_string = str_replace("	", '', $search_string);
        $search_string = strip_tags($search_string);

        curl_setopt($ch, CURLOPT_URL, $custom_vectorization_api_config->get('url'));
        curl_setopt($ch, CURLOPT_RETURNTRANSFER, 1);
        curl_setopt($ch, CURLOPT_POST, 1);
        curl_setopt($ch, CURLOPT_POSTFIELDS, "{ \"input\": \"" . $search_string . "\" }");

        $headers = [];
        $headers[] = 'Content-Type: application/json';
        $headers[] = 'Authorization: basic ' . $custom_vectorization_api_config->get('auth_token');
        curl_setopt($ch, CURLOPT_HTTPHEADER, $headers);

        $result = curl_exec($ch);
        if (!property_exists(json_decode($result), "embedding")) {
          $this->loggerFactory->get('semantic_search')->error('An exception occurred while trying to generate an embedding using the custom API.');
        }
        $embedding = json_decode($result)->embedding;
        if (curl_errno($ch)) {
          echo 'Error:' . curl_error($ch);
        }
        curl_close($ch);
        break;
    }

    return $embedding;
  }

  /**
   * Returns the vectorization service.
   *
   * @return array|mixed|null
   *   The service.
   */
  public function getVectorizationService() {
    return $this->configFactory->get('semantic_search.settings')->get('embedding_vectorization');
  }

  /**
   * Returns the current vector size.
   *
   * @return mixed
   *   The vector size.
   */
  public function getVectorSize() {
    $vectorization_service = $this->getVectorizationService();
    return self::STORAGE_OPTIONS[$vectorization_service]['vector_size'];
  }

}
